@@ -1,6 +1,5 @@
 use std::{
     borrow::Borrow,
-    collections::{HashMap, HashSet},
     path::{Path, PathBuf},
     rc::Rc,
     sync::Arc,
@@ -13,11 +12,13 @@ use exocore_chain::{
     operation::{Operation, OperationId},
     pending, EngineHandle, EngineOperationStatus,
 };
-use exocore_core::{cell::FullCell, time::Clock, time::Instant};
+use exocore_core::{
+    cell::FullCell,
+    time::{Clock, Instant},
+};
 use exocore_protos::{
     generated::exocore_store::{
-        entity_mutation::Mutation, Entity, EntityMutation, EntityQuery,
-        EntityResult as EntityResultProto, EntityResultSource, EntityResults, Trait,
+        entity_mutation::Mutation, EntityMutation, EntityQuery, EntityResults, Trait,
     },
     prost::{Message, ProstDateTimeExt},
     registry::Registry,
@@ -26,25 +27,22 @@ use exocore_protos::{
 use gc::GarbageCollector;
 use itertools::Itertools;
 
-use super::{
-    mutation_index::{IndexOperation, MutationIndex, MutationMetadata},
-    top_results::ReScoredTopResultsIterable,
-};
-use crate::{
-    entity::EntityId,
-    error::Error,
-    ordering::{OrderingValueExt, OrderingValueWrapper},
-};
+use super::mutation_index::{IndexOperation, MutationIndex, MutationMetadata};
+use crate::error::Error;
 
 mod config;
 pub use config::*;
 
 mod aggregator;
-pub(crate) use aggregator::*;
+pub use aggregator::*;
 
 mod gc;
 pub use gc::GarbageCollectorConfig;
 
+mod searcher;
+
+pub mod iterator;
+
 #[cfg(test)]
 pub(crate) mod test_index;
 #[cfg(test)]
@@ -233,197 +231,27 @@ where
         Ok((affected_operations, index_operations_count))
     }
 
-    /// Executes a search query on the indices, and returning all entities
-    /// matching the query.
+    pub fn maybe_index_chain_blocks(&mut self) -> Result<Vec<OperationId>, Error> {
+        let mut affected_operations = Vec::new();
+        self.index_chain_new_blocks(Some(&mut affected_operations))?;
+        Ok(affected_operations)
+    }
+
     pub fn search<Q: Borrow<EntityQuery>>(&self, query: Q) -> Result<EntityResults, Error> {
-        let begin_instant = Instant::now();
-
-        let query = query.borrow();
-        let query_include_deleted = query.include_deleted;
-        let mut query_page = query
-            .paging
-            .clone()
-            .unwrap_or_else(crate::query::default_paging);
-        crate::query::fill_default_paging(&mut query_page);
-
-        let reference_boost = query
-            .ordering
-            .as_ref()
-            .map_or(true, |o| !o.no_reference_boost);
-
-        // query pending & chain mutation index without original query paging since we
-        // need to do our own paging here since we are re-ranking results and
-        // that we may have more than one mutation match for each entity.
-        let mutations_query = EntityQuery {
-            paging: None,
-            ..query.clone()
-        };
-        let chain_results = self.chain_index.search_iter(&mutations_query)?;
-        let chain_hits = chain_results.total_results;
-        let pending_results = self.pending_index.search_iter(&mutations_query)?;
-        let pending_hits = pending_results.total_results;
-        let after_query_instant = Instant::now();
-
-        // create merged iterator, returning results from both underlying in order
-        let chain_results = chain_results.map(|res| (res, EntityResultSource::Chain));
-        let pending_results = pending_results.map(|res| (res, EntityResultSource::Pending));
-        let combined_results = chain_results
-            .merge_by(pending_results, |(res1, _src1), (res2, _src2)| {
-                res1.sort_value >= res2.sort_value
-            });
-
-        let hasher = result_hasher();
-        let mut digest = hasher.digest();
-        let mut entity_mutations_cache = HashMap::<EntityId, Rc<EntityAggregator>>::new();
-        let mut matched_entities = HashSet::new();
-
-        // iterate through results and returning the first N entities
-        let mut entity_results = combined_results
-            // iterate through results, starting with best scores
-            .flat_map(|(matched_mutation, index_source)| {
-                let entity_id = matched_mutation.entity_id.clone();
-
-                // check if we already processed this entity through another trait result that
-                // ranked higher
-                if matched_entities.contains(&entity_id) {
-                    return None;
-                }
-
-                // fetch all entity mutations and cache them since we may have multiple hits for
-                // same entity for traits that may have been removed since
-                let entity_mutations = self.fetch_and_cache_entity_mutations_metadata(
-                    &mut entity_mutations_cache,
-                    &entity_id,
-                    &query.projections,
-                )?;
-
-                let operation_still_present = entity_mutations
-                    .active_operations
-                    .contains(&matched_mutation.operation_id);
-                if entity_mutations.deletion_date.is_some() || !operation_still_present {
-                    // we are here if the entity has been deleted (ex: explicitly or no traits
-                    // remaining) or if the mutation metadata that was returned
-                    // by the mutation index is not active anymore, which means
-                    // that it got overridden by a subsequent operation.
-                    //
-                    // We let the garbage collector know that the entity may need to be garbage
-                    // collected. The actual garbage collection will be done
-                    // asynchronously at later time.
-                    self.gc.maybe_flag_for_collection(&entity_mutations);
-
-                    if !query_include_deleted {
-                        return None;
-                    }
-                }
-                matched_entities.insert(matched_mutation.entity_id.clone());
-
-                // Unless disabled, penalizes the entity score if it doesn't have a reference to
-                // another object
-                let mut ordering_value = matched_mutation.sort_value.clone();
-                let original_ordering_value = ordering_value.clone();
-                if reference_boost && ordering_value.is_score() && !entity_mutations.has_reference {
-                    ordering_value.boost_score(0.3);
-                };
-
-                if ordering_value.value.is_within_page_bound(&query_page) {
-                    let result = EntityResult {
-                        matched_mutation,
-                        ordering_value: ordering_value.clone(),
-                        original_ordering_value,
-                        proto: EntityResultProto {
-                            entity: Some(Entity {
-                                id: entity_id,
-                                traits: Vec::new(),
-                                creation_date: opt_date_to_proto(entity_mutations.creation_date),
-                                modification_date: opt_date_to_proto(
-                                    entity_mutations.modification_date,
-                                ),
-                                deletion_date: opt_date_to_proto(entity_mutations.deletion_date),
-                                last_operation_id: entity_mutations.last_operation_id,
-                            }),
-                            source: index_source.into(),
-                            ordering_value: Some(ordering_value.value),
-                            hash: entity_mutations.hash,
-                        },
-                        mutations: entity_mutations,
-                    };
-
-                    Some(result)
-                } else {
-                    None
-                }
-            })
-            // this steps consumes the results up until we reach the best 10 results based on the
-            // score of the highest matching trait, but re-scored negatively based on
-            // other traits
-            .top_negatively_rescored_results(query_page.count as usize, |result: &EntityResult| {
-                (
-                    result.original_ordering_value.clone(),
-                    result.ordering_value.clone(),
-                )
-            })
-            // accumulate results
-            .fold(
-                Vec::new(),
-                |mut results: Vec<EntityResult>, result: EntityResult| {
-                    digest.update(&result.mutations.hash.to_ne_bytes());
-                    results.push(result);
-                    results
-                },
-            );
-
-        let after_aggregate_instant = Instant::now();
-
-        let next_page = if let Some(last_result) = entity_results.last() {
-            let mut new_page = query_page.clone();
-
-            let ascending = query
-                .ordering
-                .as_ref()
-                .map(|s| s.ascending)
-                .unwrap_or(false);
-            if !ascending {
-                new_page.before_ordering_value = Some(last_result.ordering_value.value.clone());
-            } else {
-                new_page.after_ordering_value = Some(last_result.ordering_value.value.clone());
-            }
-
-            Some(new_page)
-        } else {
-            None
-        };
-
-        // if query specifies a `result_hash` and that new results have the same hash,
-        // we don't fetch results' data
-        let results_hash = digest.finalize();
-        let skipped_hash = results_hash == query.result_hash;
-        if !skipped_hash {
-            self.populate_results_traits(&mut entity_results, query.include_deleted);
-        }
-
-        let end_instant = Instant::now();
-        debug!(
-            "Query done chain_hits={} pending_hits={} aggr_fetch={} query={:?} aggr={:?} fetch={:?} total={:?} page={:?} next_page={:?}",
-            chain_hits,
-            pending_hits,
-            entity_mutations_cache.len(),
-            after_query_instant - begin_instant,
-            after_aggregate_instant - after_query_instant,
-            end_instant - after_aggregate_instant,
-            end_instant - begin_instant,
-            query_page,
-            next_page,
+        let searcher = searcher::Searcher::new(
+            &self.chain_index,
+            &self.pending_index,
+            &self.gc,
+            |cache, entity_id, projections| {
+                self.fetch_and_cache_entity_mutations_metadata(cache, entity_id, projections)
+            },
+            |entity_results, include_deleted| {
+                self.populate_results_traits(entity_results, include_deleted)
+            },
+            query.borrow(),
         );
 
-        let entities = entity_results.into_iter().map(|res| res.proto).collect();
-        Ok(EntityResults {
-            entities,
-            skipped_hash,
-            next_page,
-            current_page: Some(query_page),
-            estimated_count: (chain_hits + pending_hits) as u32,
-            hash: results_hash,
-        })
+        searcher.search()
     }
 
     /// Calls the garbage collector to run a pass on entities that got flagged
@@ -631,23 +459,25 @@ where
                 index_ops
             });
 
+        let before_apply = Instant::now();
         self.chain_index.apply_operations(chain_index_mutations)?;
 
         let index_operations_count = pending_index_mutations.len();
         if index_operations_count > 0 {
-            info!(
-                "Indexed in chain, and deleted from pending {} operations (from offset {:?}). New chain index last offset is {:?}.",
-                index_operations_count,
-                offset_from,
-                new_highest_block_offset
-            );
-
             self.pending_index
                 .apply_operations(pending_index_mutations.into_iter())?;
 
             if let Some(new_highest_block_offset) = new_highest_block_offset {
                 self.chain_index_last_block = Some(new_highest_block_offset);
             }
+
+            info!(
+                "Indexed in chain, and deleted from pending {} operations (from offset {:?}) in {:?}. New chain index last offset is {:?}.",
+                index_operations_count,
+                offset_from,
+                before_apply.elapsed(),
+                new_highest_block_offset
+            );
         }
 
         Ok(index_operations_count)
@@ -704,11 +534,14 @@ where
     /// Fetches an entity and all its traits from indices and the chain layer.
     /// Traits returned follow mutations in order of operation id.
     #[cfg(test)]
-    fn fetch_entity(&self, entity_id: &str) -> Result<Entity, Error> {
+    fn fetch_entity(
+        &self,
+        entity_id: &str,
+    ) -> Result<exocore_protos::generated::exocore_store::Entity, Error> {
         let aggr = self.fetch_aggregated_entity_mutations(entity_id)?;
         let traits = self.fetch_entity_traits(&aggr, false);
 
-        Ok(Entity {
+        Ok(exocore_protos::generated::exocore_store::Entity {
             id: entity_id.to_string(),
             traits,
             creation_date: aggr.creation_date.map(|t| t.to_proto_timestamp()),
@@ -720,7 +553,7 @@ where
 
     /// Fetches indexed mutations metadata from pending and chain indices for
     /// this entity id and aggregate them.
-    pub fn fetch_aggregated_entity_mutations(
+    pub(super) fn fetch_aggregated_entity_mutations(
         &self,
         entity_id: &str,
     ) -> Result<EntityAggregator, Error> {
@@ -730,7 +563,7 @@ where
 
     /// Fetches indexed mutations metadata from pending and chain indices for
     /// this entity id.
-    pub fn fetch_entity_mutations_metadata(
+    pub(super) fn fetch_entity_mutations_metadata(
         &self,
         entity_id: &str,
     ) -> Result<impl Iterator<Item = MutationMetadata>, Error> {
@@ -756,7 +589,7 @@ where
     /// Fetches and cache indexed mutations metadata.
     fn fetch_and_cache_entity_mutations_metadata(
         &self,
-        cache: &mut HashMap<String, Rc<EntityAggregator>>,
+        cache: &mut searcher::EntityMetaCache,
         entity_id: &str,
         projections: &[Projection],
     ) -> Option<Rc<EntityAggregator>> {
@@ -791,7 +624,7 @@ where
     /// from the chain layer.
     fn populate_results_traits(
         &self,
-        entity_results: &mut Vec<EntityResult>,
+        entity_results: &mut Vec<searcher::SearchResult>,
         include_deleted: bool,
     ) {
         for entity_result in entity_results {
@@ -913,19 +746,3 @@ where
         }
     }
 }
-
-/// Wrapper for entity result with matched mutation from store layer along
-/// aggregated traits.
-pub struct EntityResult {
-    pub matched_mutation: MutationMetadata,
-    pub ordering_value: OrderingValueWrapper,
-    pub original_ordering_value: OrderingValueWrapper,
-    pub proto: EntityResultProto,
-    pub mutations: Rc<EntityAggregator>,
-}
-
-fn opt_date_to_proto(
-    dt: Option<chrono::DateTime<chrono::Utc>>,
-) -> Option<exocore_protos::prost::Timestamp> {
-    dt.map(|t| t.to_proto_timestamp())
-}
