@@ -14,16 +14,12 @@ use exocore_chain::block::BlockOffset;
 use exocore_core::time::Instant;
 use exocore_protos::{
     generated::exocore_store::{
-        entity_query::Predicate, ordering, ordering_value, trait_field_predicate, trait_query,
-        EntityQuery, IdsPredicate, MatchPredicate, OperationsPredicate, Ordering, OrderingValue,
-        Paging, ReferencePredicate, TraitFieldPredicate, TraitFieldReferencePredicate,
-        TraitPredicate,
+        ordering, ordering_value, EntityQuery, Ordering, OrderingValue, Paging,
     },
     prost::{Any, ProstTimestampExt},
     reflect,
     reflect::{DynamicMessage, FieldDescriptor, FieldValue, ReflectMessage},
     registry::Registry,
-    store::AllPredicate,
 };
 pub use operations::*;
 pub use results::*;
@@ -31,12 +27,16 @@ use tantivy::{
     collector::{Collector, Count, MultiCollector, TopDocs},
     directory::MmapDirectory,
     fastfield::FastFieldReader,
-    query::{AllQuery, BooleanQuery, FuzzyTermQuery, Occur, PhraseQuery, Query, TermQuery},
+    query::{AllQuery, TermQuery},
     schema::{Field, IndexRecordOption},
     DocAddress, Document, Index as TantivyIndex, IndexReader, IndexSettings, IndexSortByField,
     IndexWriter, Order, ReloadPolicy, Searcher, SegmentReader, Term,
 };
 
+use self::{
+    query::{ParsedQuery, QueryParser},
+    schema::MutationIndexSchema,
+};
 use crate::{
     entity::EntityIdRef, error::Error, mutation::OperationId, ordering::OrderingValueWrapper,
 };
@@ -44,6 +44,7 @@ use crate::{
 mod config;
 mod entity_cache;
 mod operations;
+mod query;
 mod results;
 mod schema;
 #[cfg(test)]
@@ -63,11 +64,11 @@ pub struct MutationIndex {
     index: TantivyIndex,
     index_reader: IndexReader,
     index_writer: Mutex<IndexWriter>,
-    schemas: Arc<Registry>,
-    fields: schema::Fields,
+    schema: MutationIndexSchema,
     storage: Storage,
     entity_cache: EntityMutationsCache,
     full_text_boost: f32,
+    schema_registry: Arc<Registry>,
 }
 
 #[derive(Debug, PartialEq, Eq)]
@@ -80,17 +81,17 @@ impl MutationIndex {
     /// Creates or opens a disk persisted index.
     pub fn open_or_create_mmap(
         config: MutationIndexConfig,
-        schemas: Arc<Registry>,
+        schema_registry: Arc<Registry>,
         directory: &Path,
     ) -> Result<MutationIndex, Error> {
-        let (tantivy_schema, fields) = schema::build_tantivy_schema(config, schemas.as_ref());
+        let schema = MutationIndexSchema::new(config, schema_registry.as_ref());
 
         let directory = MmapDirectory::open(directory)?;
         let index = TantivyIndex::builder()
-            .schema(tantivy_schema)
+            .schema(schema.tantivy.clone())
             .settings(index_settings())
             .open_or_create(directory)?;
-        fields.register_tokenizers(&index);
+        schema.register_tokenizers(&index);
 
         let index_reader = index
             .reader_builder()
@@ -108,26 +109,26 @@ impl MutationIndex {
             index,
             index_reader,
             index_writer: Mutex::new(index_writer),
-            schemas,
-            fields,
+            schema,
             storage: Storage::Disk,
             entity_cache: EntityMutationsCache::new(config.entity_mutations_cache_size as usize),
             full_text_boost: 1.0,
+            schema_registry,
         })
     }
 
     /// Creates or opens a in-memory index.
     pub fn create_in_memory(
         config: MutationIndexConfig,
-        schemas: Arc<Registry>,
+        schema_registry: Arc<Registry>,
     ) -> Result<MutationIndex, Error> {
-        let (tantivy_schema, fields) = schema::build_tantivy_schema(config, schemas.as_ref());
+        let schema = MutationIndexSchema::new(config, schema_registry.as_ref());
 
         let index = TantivyIndex::builder()
-            .schema(tantivy_schema)
+            .schema(schema.tantivy.clone())
             .settings(index_settings())
             .create_in_ram()?;
-        fields.register_tokenizers(&index);
+        schema.register_tokenizers(&index);
 
         let index_reader = index
             .reader_builder()
@@ -145,11 +146,11 @@ impl MutationIndex {
             index,
             index_reader,
             index_writer: Mutex::new(index_writer),
-            schemas,
-            fields,
+            schema,
             storage: Storage::Memory,
             entity_cache: EntityMutationsCache::new(config.entity_mutations_cache_size as usize),
             full_text_boost: 1.0,
+            schema_registry,
         })
     }
 
@@ -243,7 +244,7 @@ impl MutationIndex {
                     self.entity_cache.remove(&entity_id);
 
                     index_writer
-                        .delete_term(Term::from_field_u64(self.fields.operation_id, operation_id));
+                        .delete_term(Term::from_field_u64(self.schema.operation_id, operation_id));
                 }
             }
 
@@ -275,7 +276,7 @@ impl MutationIndex {
         let searcher = self.index_reader.searcher();
 
         let query = AllQuery;
-        let top_collector = TopDocs::with_limit(1).order_by_u64_field(self.fields.block_offset);
+        let top_collector = TopDocs::with_limit(1).order_by_u64_field(self.schema.block_offset);
         let search_results = searcher.search(&query, &top_collector)?;
 
         Ok(search_results
@@ -286,24 +287,11 @@ impl MutationIndex {
     /// Execute a query on the index and return a page of mutations matching the
     /// query.
     pub fn search<Q: Borrow<EntityQuery>>(&self, query: Q) -> Result<MutationResults, Error> {
-        let query = query.borrow();
-        let predicate = query
-            .predicate
-            .as_ref()
-            .ok_or(Error::ProtoFieldExpected("predicate"))?;
+        let parsed_query =
+            QueryParser::parse(&self.index, &self.schema, &self.config, query.borrow())?;
 
-        let paging = query.paging.as_ref();
-        let ordering = query.ordering.as_ref();
-
-        let results = match &predicate {
-            Predicate::Trait(inner) => self.search_with_trait(inner, paging, ordering),
-            Predicate::Match(inner) => self.search_matches(inner, paging, ordering),
-            Predicate::Ids(inner) => self.search_entity_ids(inner, paging, ordering),
-            Predicate::Reference(inner) => self.search_reference(inner, paging, ordering),
-            Predicate::Operations(inner) => self.search_operations(inner, paging, ordering),
-            Predicate::All(inner) => self.search_all(inner, paging, ordering),
-            Predicate::Test(_inner) => Err(anyhow!("Query failed for tests").into()),
-        }?;
+        let searcher = self.index_reader.searcher();
+        let results = self.execute_tantivy_query_with_paging(searcher, parsed_query)?;
 
         Ok(results)
     }
@@ -342,8 +330,8 @@ impl MutationIndex {
 
         let searcher = self.index_reader.searcher();
 
-        let term = Term::from_field_text(self.fields.entity_id, entity_id);
-        let query = TermQuery::new(term, IndexRecordOption::Basic);
+        let term = Term::from_field_text(self.schema.entity_id, entity_id);
+        let tantivy = TermQuery::new(term, IndexRecordOption::Basic);
 
         let ordering = Ordering {
             ascending: true,
@@ -355,13 +343,14 @@ impl MutationIndex {
             ..Default::default()
         };
 
-        let mut results = self.execute_tantivy_query_with_paging(
-            searcher,
-            &query,
-            Some(&paging),
+        let parsed_query = ParsedQuery {
+            tantivy: Box::new(tantivy),
+            paging,
             ordering,
-            None,
-        )?;
+            trait_name: None,
+        };
+
+        let mut results = self.execute_tantivy_query_with_paging(searcher, parsed_query)?;
 
         // because of the way we index pending (we may have pending store events after
         // indexing it after first), we need to make sure we don't include any
@@ -376,168 +365,6 @@ impl MutationIndex {
         Ok(entity_mutations)
     }
 
-    /// Execute a search by trait type query and return traits in operations id
-    /// descending order.
-    fn search_with_trait(
-        &self,
-        predicate: &TraitPredicate,
-        paging: Option<&Paging>,
-        ordering: Option<&Ordering>,
-    ) -> Result<MutationResults, Error> {
-        let searcher = self.index_reader.searcher();
-
-        let mut ordering = ordering.cloned().unwrap_or_default();
-        if ordering.value.is_none() {
-            ordering.value = Some(ordering::Value::OperationId(true));
-        }
-
-        let mut queries: Vec<(Occur, Box<dyn Query>)> = Vec::new();
-
-        let trait_type = Term::from_field_text(self.fields.trait_type, &predicate.trait_name);
-        let trait_type_query = TermQuery::new(trait_type, IndexRecordOption::Basic);
-        queries.push((Occur::Must, Box::new(trait_type_query)));
-
-        if let Some(trait_query) = &predicate.query {
-            match &trait_query.predicate {
-                Some(trait_query::Predicate::Match(trait_pred)) => {
-                    queries.push((Occur::Must, self.match_predicate_to_query(trait_pred)?));
-                }
-                Some(trait_query::Predicate::Field(trait_pred)) => {
-                    queries.push((
-                        Occur::Must,
-                        self.trait_field_predicate_to_query(&predicate.trait_name, trait_pred)?,
-                    ));
-                }
-                Some(trait_query::Predicate::Reference(trait_pred)) => {
-                    queries.push((
-                        Occur::Must,
-                        self.trait_field_reference_predicate_to_query(
-                            &predicate.trait_name,
-                            trait_pred,
-                        )?,
-                    ));
-                }
-                None => {}
-            }
-        }
-
-        let query = BooleanQuery::from(queries);
-        self.execute_tantivy_query_with_paging(
-            searcher,
-            &query,
-            paging,
-            ordering,
-            Some(&predicate.trait_name),
-        )
-    }
-
-    /// Execute a search by text query
-    fn search_matches(
-        &self,
-        predicate: &MatchPredicate,
-        paging: Option<&Paging>,
-        ordering: Option<&Ordering>,
-    ) -> Result<MutationResults, Error> {
-        let searcher = self.index_reader.searcher();
-
-        let mut ordering = ordering.cloned().unwrap_or_default();
-        if ordering.value.is_none() {
-            ordering.value = Some(ordering::Value::Score(true));
-        }
-
-        let query = self.match_predicate_to_query(predicate)?;
-        self.execute_tantivy_query_with_paging(searcher, &query, paging, ordering, None)
-    }
-
-    /// Executes a search for mutations with the given operations ids.
-    pub fn search_operations(
-        &self,
-        predicate: &OperationsPredicate,
-        paging: Option<&Paging>,
-        ordering: Option<&Ordering>,
-    ) -> Result<MutationResults, Error> {
-        let searcher = self.index_reader.searcher();
-
-        let mut queries: Vec<(Occur, Box<dyn Query>)> = Vec::new();
-        for operation_id in &predicate.operation_ids {
-            let op_term = Term::from_field_u64(self.fields.operation_id, *operation_id);
-            let op_query = TermQuery::new(op_term, IndexRecordOption::Basic);
-            queries.push((Occur::Should, Box::new(op_query)));
-        }
-        let query = BooleanQuery::from(queries);
-
-        let mut ordering = ordering.cloned().unwrap_or_default();
-        if ordering.value.is_none() {
-            ordering.value = Some(ordering::Value::OperationId(true));
-            ordering.ascending = true;
-        }
-
-        self.execute_tantivy_query_with_paging(searcher, &query, paging, ordering, None)
-    }
-
-    /// Executes a search for mutations on the given entities ids.
-    pub fn search_entity_ids(
-        &self,
-        predicate: &IdsPredicate,
-        paging: Option<&Paging>,
-        ordering: Option<&Ordering>,
-    ) -> Result<MutationResults, Error> {
-        let searcher = self.index_reader.searcher();
-
-        let mut queries: Vec<(Occur, Box<dyn Query>)> = Vec::new();
-        for entity_id in &predicate.ids {
-            let term = Term::from_field_text(self.fields.entity_id, entity_id);
-            let query = TermQuery::new(term, IndexRecordOption::Basic);
-            queries.push((Occur::Should, Box::new(query)));
-        }
-        let query = BooleanQuery::from(queries);
-
-        let mut ordering = ordering.cloned().unwrap_or_default();
-        if ordering.value.is_none() {
-            ordering.value = Some(ordering::Value::OperationId(true));
-        }
-
-        self.execute_tantivy_query_with_paging(searcher, &query, paging, ordering, None)
-    }
-
-    /// Executes a search for traits that have the given reference to another
-    /// entity and optionally trait.
-    fn search_reference(
-        &self,
-        predicate: &ReferencePredicate,
-        paging: Option<&Paging>,
-        ordering: Option<&Ordering>,
-    ) -> Result<MutationResults, Error> {
-        let searcher = self.index_reader.searcher();
-
-        let query = self.reference_predicate_to_query(self.fields.all_refs, predicate);
-
-        let mut ordering = ordering.cloned().unwrap_or_default();
-        if ordering.value.is_none() {
-            ordering.value = Some(ordering::Value::OperationId(true));
-        }
-
-        self.execute_tantivy_query_with_paging(searcher, &query, paging, ordering, None)
-    }
-
-    /// Returns all mutations.
-    pub fn search_all(
-        &self,
-        _predicate: &AllPredicate,
-        paging: Option<&Paging>,
-        ordering: Option<&Ordering>,
-    ) -> Result<MutationResults, Error> {
-        let searcher = self.index_reader.searcher();
-
-        let mut ordering = ordering.cloned().unwrap_or_default();
-        if ordering.value.is_none() {
-            ordering.value = Some(ordering::Value::OperationId(true));
-            ordering.ascending = false;
-        }
-
-        self.execute_tantivy_query_with_paging(searcher, &AllQuery, paging, ordering, None)
-    }
-
     /// Converts a trait put / update to Tantivy document
     fn trait_put_to_document(&self, operation: &PutTraitMutation) -> Result<Document, Error> {
         let message_any = operation
@@ -549,33 +376,33 @@ impl MutationIndex {
         let mut doc = Document::default();
 
         let message_full_name = reflect::any_url_to_full_name(&message_any.type_url);
-        doc.add_text(self.fields.trait_type, &message_full_name);
-        doc.add_text(self.fields.trait_id, &operation.trt.id);
-        doc.add_text(self.fields.entity_id, &operation.entity_id);
+        doc.add_text(self.schema.trait_type, &message_full_name);
+        doc.add_text(self.schema.trait_id, &operation.trt.id);
+        doc.add_text(self.schema.entity_id, &operation.entity_id);
         doc.add_text(
-            self.fields.entity_trait_id,
+            self.schema.entity_trait_id,
             &format!("{}_{}", operation.entity_id, &operation.trt.id),
         );
 
-        doc.add_u64(self.fields.operation_id, operation.operation_id);
+        doc.add_u64(self.schema.operation_id, operation.operation_id);
         if let Some(block_offset) = operation.block_offset {
-            doc.add_u64(self.fields.block_offset, block_offset);
+            doc.add_u64(self.schema.block_offset, block_offset);
         }
 
         if let Some(creation_date) = &operation.trt.creation_date {
             doc.add_u64(
-                self.fields.creation_date,
+                self.schema.creation_date,
                 creation_date.to_timestamp_nanos(),
             );
         }
         if let Some(modification_date) = &operation.trt.modification_date {
             doc.add_u64(
-                self.fields.modification_date,
+                self.schema.modification_date,
                 modification_date.to_timestamp_nanos(),
             );
         }
 
-        doc.add_u64(self.fields.document_type, MutationType::TRAIT_PUT_ID);
+        doc.add_u64(self.schema.document_type, MutationType::TRAIT_PUT_ID);
 
         self.trait_message_to_document(&mut doc, message_any, message_full_name);
 
@@ -590,7 +417,8 @@ impl MutationIndex {
         message_any: &Any,
         message_full_name: String,
     ) {
-        let dyn_message = match reflect::from_prost_any(self.schemas.as_ref(), message_any) {
+        let dyn_message = match reflect::from_prost_any(self.schema_registry.as_ref(), message_any)
+        {
             Ok(dyn_message) => dyn_message,
             Err(err) => {
                 error!(
@@ -601,7 +429,7 @@ impl MutationIndex {
         };
 
         let message_mappings = if let Some(message_mappings) =
-            self.fields.dynamic_mappings.get(dyn_message.full_name())
+            self.schema.dynamic_fields.get(dyn_message.full_name())
         {
             message_mappings
         } else {
@@ -625,7 +453,7 @@ impl MutationIndex {
         // the message has at least one reference field
         if has_reference {
             doc.add_u64(
-                self.fields.has_reference,
+                self.schema.has_reference,
                 schema::bool_to_u64(has_reference),
             );
         }
@@ -658,7 +486,7 @@ impl MutationIndex {
         // they don't have a mapping directly since mapping is done on
         // sub-fields (ex: mapped into `field.sub_field`)
         if let FieldValue::Message(_, _) = &field_value {
-            match field_value.into_message(&self.schemas) {
+            match field_value.into_message(&self.schema_registry) {
                 Ok(dyn_msg) => {
                     for sub_field in dyn_msg.fields().values() {
                         self.add_trait_message_document_field(
@@ -688,7 +516,7 @@ impl MutationIndex {
         match field_value {
             FieldValue::String(value) if field_desc.text_flag => {
                 doc.add_text(mapped_field.field, &value);
-                doc.add_text(self.fields.all_text, &value);
+                doc.add_text(self.schema.all_text, &value);
             }
             FieldValue::String(value) if field_desc.indexed_flag => {
                 doc.add_text(mapped_field.field, &value);
@@ -696,7 +524,7 @@ impl MutationIndex {
             FieldValue::Reference(value) if field_desc.indexed_flag => {
                 let ref_value = format!("entity{} trait{}", value.entity_id, value.trait_id);
                 doc.add_text(mapped_field.field, &ref_value);
-                doc.add_text(self.fields.all_refs, &ref_value);
+                doc.add_text(self.schema.all_refs, &ref_value);
                 *has_reference = true;
             }
             FieldValue::DateTime(value) if field_desc.indexed_flag || field_desc.sorted_flag => {
@@ -727,19 +555,19 @@ impl MutationIndex {
     fn trait_tombstone_to_document(&self, operation: &PutTraitTombstoneMutation) -> Document {
         let mut doc = Document::default();
 
-        doc.add_text(self.fields.trait_id, &operation.trait_id);
-        doc.add_text(self.fields.entity_id, &operation.entity_id);
+        doc.add_text(self.schema.trait_id, &operation.trait_id);
+        doc.add_text(self.schema.entity_id, &operation.entity_id);
         doc.add_text(
-            self.fields.entity_trait_id,
+            self.schema.entity_trait_id,
             &format!("{}_{}", operation.entity_id, operation.trait_id),
         );
-        doc.add_u64(self.fields.operation_id, operation.operation_id);
+        doc.add_u64(self.schema.operation_id, operation.operation_id);
 
         if let Some(block_offset) = operation.block_offset {
-            doc.add_u64(self.fields.block_offset, block_offset);
+            doc.add_u64(self.schema.block_offset, block_offset);
         }
 
-        doc.add_u64(self.fields.document_type, MutationType::TRAIT_TOMBSTONE_ID);
+        doc.add_u64(self.schema.document_type, MutationType::TRAIT_TOMBSTONE_ID);
 
         doc
     }
@@ -748,14 +576,14 @@ impl MutationIndex {
     fn entity_tombstone_to_document(&self, operation: &PutEntityTombstoneMutation) -> Document {
         let mut doc = Document::default();
 
-        doc.add_text(self.fields.entity_id, &operation.entity_id);
-        doc.add_u64(self.fields.operation_id, operation.operation_id);
+        doc.add_text(self.schema.entity_id, &operation.entity_id);
+        doc.add_u64(self.schema.operation_id, operation.operation_id);
 
         if let Some(block_offset) = operation.block_offset {
-            doc.add_u64(self.fields.block_offset, block_offset);
+            doc.add_u64(self.schema.block_offset, block_offset);
         }
 
-        doc.add_u64(self.fields.document_type, MutationType::ENTITY_TOMBSTONE_ID);
+        doc.add_u64(self.schema.document_type, MutationType::ENTITY_TOMBSTONE_ID);
 
         doc
     }
@@ -770,181 +598,85 @@ impl MutationIndex {
     ) -> Document {
         let mut doc = Document::default();
 
-        doc.add_text(self.fields.entity_id, entity_id);
-        doc.add_u64(self.fields.operation_id, operation_id);
-        doc.add_u64(self.fields.document_type, MutationType::PENDING_DELETION_ID);
+        doc.add_text(self.schema.entity_id, entity_id);
+        doc.add_u64(self.schema.operation_id, operation_id);
+        doc.add_u64(self.schema.document_type, MutationType::PENDING_DELETION_ID);
 
         doc
     }
 
-    /// Transforms a text match predicate to Tantivy query.
-    fn match_predicate_to_query(
-        &self,
-        predicate: &MatchPredicate,
-    ) -> Result<Box<dyn tantivy::query::Query>, Error> {
-        let field = self.fields.all_text;
-        let text = predicate.query.as_str();
-        let no_fuzzy = predicate.no_fuzzy;
-        Ok(Box::new(self.new_fuzzy_query(field, text, no_fuzzy)?))
-    }
-
-    /// Create a fuzzy query for a field and given text.
-    fn new_fuzzy_query(
-        &self,
-        field: Field,
-        text: &str,
-        no_fuzzy: bool,
-    ) -> Result<BooleanQuery, Error> {
-        let tok = self.index.tokenizer_for_field(field)?;
-        let mut queries: Vec<(Occur, Box<dyn Query>)> = Vec::new();
-        let mut stream = tok.token_stream(text);
-
-        while stream.advance() {
-            let token = stream.token().text.as_str();
-            let term = Term::from_field_text(field, token);
-
-            if !no_fuzzy && token.len() > 3 {
-                let max_distance = if token.len() > 6 { 2 } else { 1 };
-                let query = Box::new(FuzzyTermQuery::new(term.clone(), max_distance, true));
-                queries.push((Occur::Should, query));
-            }
-
-            // even if fuzzy is enabled, we add the term again so that an exact match scores
-            // more
-            let query = Box::new(TermQuery::new(
-                term,
-                IndexRecordOption::WithFreqsAndPositions,
-            ));
-            queries.push((Occur::Should, query));
-        }
-
-        Ok(BooleanQuery::from(queries))
-    }
-
-    /// Transforms a trait's field predicate to Tantivy query.
-    fn trait_field_predicate_to_query(
-        &self,
-        trait_name: &str,
-        predicate: &TraitFieldPredicate,
-    ) -> Result<Box<dyn tantivy::query::Query>, Error> {
-        use reflect::FieldType as FT;
-        use trait_field_predicate::Value as PV;
-
-        let fields = self
-            .fields
-            .get_dynamic_trait_field_prefix(trait_name, &predicate.field)?;
-
-        let mut queries: Vec<(Occur, Box<dyn Query>)> = Vec::new();
-        for field in fields {
-            match (&field.field_type, &predicate.value) {
-                (FT::String, Some(PV::String(value))) => {
-                    let term = Term::from_field_text(field.field, value);
-
-                    queries.push((Occur::Should, Box::new(TermQuery::new(term, IndexRecordOption::Basic))));
-                }
-                (ft, pv) => {
-                    return Err(
-                        Error::QueryParsing(
-                            anyhow!(
-                                "Incompatible field type vs field value in predicate: trait_name={} field={}, field_type={:?}, value={:?}",
-                                trait_name,
-                                predicate.field,
-                                ft,
-                                pv,
-                            ))
-                    )
-                }
-            }
-        }
-
-        Ok(Box::new(BooleanQuery::from(queries)))
-    }
-
-    /// Transforms a trait's field reference predicate to Tantivy query.
-    fn trait_field_reference_predicate_to_query(
-        &self,
-        trait_name: &str,
-        predicate: &TraitFieldReferencePredicate,
-    ) -> Result<Box<dyn tantivy::query::Query>, Error> {
-        let field = self
-            .fields
-            .get_dynamic_trait_field(trait_name, &predicate.field)?;
-
-        let reference = predicate
-            .reference
-            .as_ref()
-            .ok_or(Error::ProtoFieldExpected("reference"))?;
-
-        Ok(self.reference_predicate_to_query(field.field, reference))
-    }
-
-    /// Transforms a reference predicate to Tantivy query.
-    fn reference_predicate_to_query(
-        &self,
-        field: Field,
-        predicate: &ReferencePredicate,
-    ) -> Box<dyn tantivy::query::Query> {
-        let query: Box<dyn tantivy::query::Query> = if !predicate.trait_id.is_empty() {
-            let terms = vec![
-                Term::from_field_text(field, &format!("entity{}", predicate.entity_id)),
-                Term::from_field_text(field, &format!("trait{}", predicate.trait_id)),
-            ];
-            Box::new(PhraseQuery::new(terms))
-        } else {
-            Box::new(TermQuery::new(
-                Term::from_field_text(field, &format!("entity{}", predicate.entity_id)),
-                IndexRecordOption::Basic,
-            ))
-        };
-        query
-    }
-
     /// Execute query on Tantivy index by taking paging, ordering into
     /// consideration and returns paged results.
     fn execute_tantivy_query_with_paging<S>(
         &self,
         searcher: S,
-        query: &dyn tantivy::query::Query,
-        paging: Option<&Paging>,
-        ordering: Ordering,
-        trait_name: Option<&str>,
+        query: ParsedQuery,
     ) -> Result<MutationResults, Error>
     where
         S: Deref<Target = Searcher>,
     {
-        let paging = paging.cloned().unwrap_or(Paging {
-            after_ordering_value: None,
-            before_ordering_value: None,
-            count: self.config.iterator_page_size,
-            offset: 0,
-        });
-
-        let ordering_value = ordering
+        let ordering_value = query
+            .ordering
             .value
             .ok_or(Error::ProtoFieldExpected("ordering.value"))?;
         let (mutations, total) = match ordering_value {
             ordering::Value::Score(_) => {
                 let collector = self.match_score_collector(
-                    &paging,
-                    ordering.ascending,
-                    ordering.no_recency_boost,
+                    &query.paging,
+                    query.ordering.ascending,
+                    query.ordering.no_recency_boost,
                 );
-                self.execute_tantity_query_with_collector(searcher, query, collector)?
+                self.execute_tantity_query_with_collector(
+                    searcher,
+                    query.tantivy.as_ref(),
+                    collector,
+                )?
             }
             ordering::Value::OperationId(_) => {
-                let sort_field = self.fields.operation_id;
-                let collector =
-                    self.sorted_field_collector(&paging, sort_field, ordering.ascending);
-                self.execute_tantity_query_with_collector(searcher, query, collector)?
+                let sort_field = self.schema.operation_id;
+                let collector = self.sorted_field_collector(
+                    &query.paging,
+                    sort_field,
+                    query.ordering.ascending,
+                );
+                self.execute_tantity_query_with_collector(
+                    searcher,
+                    query.tantivy.as_ref(),
+                    collector,
+                )?
+            }
+            ordering::Value::CreatedAt(_) => {
+                let collector = self.sorted_field_collector(
+                    &query.paging,
+                    self.schema.creation_date,
+                    query.ordering.ascending,
+                );
+                self.execute_tantity_query_with_collector(
+                    searcher,
+                    query.tantivy.as_ref(),
+                    collector,
+                )?
+            }
+            ordering::Value::UpdatedAt(_) => {
+                let collector = self.sorted_field_collector(
+                    &query.paging,
+                    self.schema.modification_date,
+                    query.ordering.ascending,
+                );
+                self.execute_tantity_query_with_collector(
+                    searcher,
+                    query.tantivy.as_ref(),
+                    collector,
+                )?
             }
             ordering::Value::Field(field_name) => {
-                let trait_name = trait_name.ok_or_else(|| {
+                let trait_name = query.trait_name.ok_or_else(|| {
                     Error::QueryParsing(anyhow!("Ordering by field only supported in trait query",))
                 })?;
 
                 let sort_field = self
-                    .fields
-                    .get_dynamic_trait_field(trait_name, &field_name)?;
+                    .schema
+                    .get_dynamic_trait_field(&trait_name, &field_name)?;
                 if !sort_field.is_fast_field {
                     return Err(Error::QueryParsing(anyhow!(
                         "Cannot sort by field '{}' as it's not sortable in  trait '{}'",
@@ -953,23 +685,30 @@ impl MutationIndex {
                     )));
                 }
 
-                let collector =
-                    self.sorted_field_collector(&paging, sort_field.field, ordering.ascending);
-                self.execute_tantity_query_with_collector(searcher, query, collector)?
+                let collector = self.sorted_field_collector(
+                    &query.paging,
+                    sort_field.field,
+                    query.ordering.ascending,
+                );
+                self.execute_tantity_query_with_collector(
+                    searcher,
+                    query.tantivy.as_ref(),
+                    collector,
+                )?
             }
         };
 
-        let next_page = if mutations.len() >= paging.count as usize {
+        let next_page = if mutations.len() >= query.paging.count as usize {
             Some(Paging {
-                count: paging.count,
-                offset: paging.offset + mutations.len() as u32,
+                count: query.paging.count,
+                offset: query.paging.offset + mutations.len() as u32,
                 ..Default::default()
             })
         } else {
             None
         };
 
-        let remaining = total - paging.offset as usize - mutations.len();
+        let remaining = total - query.paging.offset as usize - mutations.len();
         Ok(MutationResults {
             mutations,
             total,
@@ -1000,24 +739,24 @@ impl MutationIndex {
             // ignored results were out of the requested paging
             if !sort_value.ignore {
                 let doc = searcher.doc(doc_addr)?;
-                let block_offset = schema::get_doc_opt_u64_value(&doc, self.fields.block_offset);
-                let operation_id = schema::get_doc_u64_value(&doc, self.fields.operation_id);
-                let entity_id = schema::get_doc_string_value(&doc, self.fields.entity_id);
-                let opt_trait_id = schema::get_doc_opt_string_value(&doc, self.fields.trait_id);
-                let document_type_id = schema::get_doc_u64_value(&doc, self.fields.document_type);
+                let block_offset = schema::get_doc_opt_u64_value(&doc, self.schema.block_offset);
+                let operation_id = schema::get_doc_u64_value(&doc, self.schema.operation_id);
+                let entity_id = schema::get_doc_string_value(&doc, self.schema.entity_id);
+                let opt_trait_id = schema::get_doc_opt_string_value(&doc, self.schema.trait_id);
+                let document_type_id = schema::get_doc_u64_value(&doc, self.schema.document_type);
 
                 let mut mutation_type = MutationType::new(document_type_id, opt_trait_id)?;
                 if let MutationType::TraitPut(put_trait) = &mut mutation_type {
                     put_trait.creation_date =
-                        schema::get_doc_opt_u64_value(&doc, self.fields.creation_date)
+                        schema::get_doc_opt_u64_value(&doc, self.schema.creation_date)
                             .map(|ts| Utc.timestamp_nanos(ts as i64));
                     put_trait.modification_date =
-                        schema::get_doc_opt_u64_value(&doc, self.fields.modification_date)
+                        schema::get_doc_opt_u64_value(&doc, self.schema.modification_date)
                             .map(|ts| Utc.timestamp_nanos(ts as i64));
                     put_trait.trait_type =
-                        schema::get_doc_opt_string_value(&doc, self.fields.trait_type);
+                        schema::get_doc_opt_string_value(&doc, self.schema.trait_type);
                     put_trait.has_reference =
-                        schema::get_doc_opt_bool_value(&doc, self.fields.has_reference)
+                        schema::get_doc_opt_bool_value(&doc, self.schema.has_reference)
                             .unwrap_or(false);
                 }
 
@@ -1045,7 +784,7 @@ impl MutationIndex {
         sort_field: Field,
         ascending: bool,
     ) -> impl Collector<Fruit = Vec<(OrderingValueWrapper, DocAddress)>> {
-        let operation_id_field = self.fields.operation_id;
+        let operation_id_field = self.schema.operation_id;
         TopDocs::with_limit(paging.count as usize)
             .and_offset(paging.offset as usize)
             .custom_score(move |segment_reader: &SegmentReader| {
@@ -1078,8 +817,8 @@ impl MutationIndex {
         no_recency_boost: bool,
     ) -> impl Collector<Fruit = Vec<(OrderingValueWrapper, DocAddress)>> {
         let now = Utc::now().timestamp_nanos() as u64;
-        let operation_id_field = self.fields.operation_id;
-        let modification_date_field = self.fields.modification_date;
+        let operation_id_field = self.schema.operation_id;
+        let modification_date_field = self.schema.modification_date;
         let boost = self.full_text_boost;
         TopDocs::with_limit(paging.count as usize)
             .and_offset(paging.offset as usize)
